==> welltested-2.0.0.tar.gz: lib/cli/operation.dart <==
import 'dart:async';
import 'dart:io';

import 'package:analyzer/dart/analysis/analysis_context_collection.dart';
import 'package:analyzer/dart/analysis/results.dart';
import 'package:analyzer/dart/ast/ast.dart';
import 'package:analyzer/dart/element/element.dart';
import 'package:analyzer/file_system/physical_file_system.dart';
import 'package:azure_application_insights/azure_application_insights.dart';
import 'package:welltested/cli/ast_visitor.dart';
import 'package:welltested/generators/saver.dart';
import 'package:welltested/src/extensions/class_element_extension.dart';
import 'package:welltested/src/extensions/file_extensions.dart';
import 'package:welltested/src/models/pubspec_data.dart';
import 'package:welltested/src/models/test_run_info.dart';
import 'package:welltested/src/utilities/logging.dart';
import 'package:welltested/src/utilities/path_utils.dart';
import 'package:welltested/src/utilities/warnings.dart';
import 'package:welltested/src/utilities/telemetry_client.dart';

import '../generators/unit_generator.dart';
import '../src/repositories/dart_server_repository.dart';
import '../src/utilities/utilities.dart';
import '../src/visitors/class_visitor.dart';

class CLIOperation {
  static Future<void> saveUnit({required PubspecData pubspecData}) async {
    wtLog.startSpinner("Getting ready...");
    final dartServerRepository = DartServerRepository();
    await dartServerRepository.initialize();

    Iterable<FileSystemEntity> fileList = Directory(PathUtils.libPath)
        .listSync(recursive: true)
        .where((element) => element.path.endsWith(".dart"));

    AnalysisContextCollection collection = AnalysisContextCollection(
      includedPaths: [PathUtils.libPath],
      resourceProvider: PhysicalResourceProvider.INSTANCE,
    );

    List<Future> saveFutures = [];
    final testRunInfo = TestRunInfo.newRun();
    for (var file in fileList) {
      saveFutures.add(
        runSaveUnit(
            file, collection, pubspecData, dartServerRepository, testRunInfo),
      );
    }
    wtLog.startSpinner("Saving Unit Tests");
    await Future.wait(saveFutures);
    wtLog.stopSpinner(
        message:
            '✅ Succeeded. ${testRunInfo.updated} were saved.\nNote: Saving working tests helps improve future test generations. Keep it up!');
    dartServerRepository.dispose();
  }

  static Future<void> generateUnit({
    required PubspecData pubspecData,
    String? filePath,
    String? className,
    String? methodName,
  }) async {
    wtLog.startSpinner("Getting ready...");
    final dartServerRepository = DartServerRepository();
    await dartServerRepository.initialize();

    Iterable<FileSystemEntity> fileList = filePath != null
        ? [File(filePath)]
        : (Directory(PathUtils.libPath).listSync(recursive: true))
            .where((element) => element.path.endsWith(".dart"));

    AnalysisContextCollection collection = AnalysisContextCollection(
      includedPaths: [PathUtils.libPath],
      resourceProvider: PhysicalResourceProvider.INSTANCE,
    );
    List<Future> generationFutures = [];
    final testRunInfo = TestRunInfo.newRun();

    for (var file in fileList) {
      generationFutures.add(
        runUnit(
            file, collection, pubspecData, dartServerRepository, testRunInfo,
            className: className, methodName: methodName),
      );
    }
    wtLog.startSpinner('Generating unit tests');
    await Future.wait(generationFutures);

    wtLog.stopSpinner(
        message:
            '✅ Succeeded. Please find the generated tests in `/test` folder.\nCreated: ${testRunInfo.created}\nUpdated:${testRunInfo.updated}\nFailed:${testRunInfo.failed}');

    dartServerRepository.dispose();
    await runMockitoCommand();
  }
}

Future runUnit(
    FileSystemEntity file,
    AnalysisContextCollection collection,
    PubspecData pubspecData,
    DartServerRepository dartServerRepository,
    TestRunInfo testRunInfo,
    {String? className,
    String? methodName}) async {
  try {
    final astVisitor = WelltestedASTVisitor();
    final resolvedUnit = await file.resolveFile(collection);
    CompilationUnit unit = resolvedUnit.unit;
    unit.visitChildren(astVisitor);

    List<Future> generationFutures = [];

    for (final declaration in unit.declarations) {
      if (declaration is ClassDeclaration) {
        if (className != null && declaration.name.toString() != className)
          continue;
        final element = declaration.declaredElement!;
        if (!isAnnotationPresent(element)) continue;
        if (element.checkIfAWidgetOrState) {
          wtLog.warning(Warnings.annotationOnWidget(element.displayName));
        }
        generationFutures.add(UnitGenerator().generateForAnnotatedElement(
          element: element,
          annotation: element.metadata.firstWhere(
              (element) => element.element!.displayName == "Welltested"),
          pubspecData: pubspecData,
          dartServerRepository: dartServerRepository,
          classDeclaration: declaration,
          testRunInfo: testRunInfo,
          classFile: file as File,
          requestedFunction: methodName,
        ));
      }
    }
    await Future.wait(generationFutures);
  } catch (e, stackTrace) {
    wtTelemetry.trackError(
        severity: Severity.error, error: e, stackTrace: stackTrace);
  }
}

Future runSaveUnit(
  FileSystemEntity file,
  AnalysisContextCollection collection,
  PubspecData pubspecData,
  DartServerRepository dartServerRepository,
  TestRunInfo testRunInfo,
) async {
  try {
    final astVisitor = WelltestedASTVisitor();
    final resolvedUnit = await file.resolveFile(collection);
    CompilationUnit unit = resolvedUnit.unit;
    unit.visitChildren(astVisitor);

    List<Future> generationFutures = [];

    for (final declaration in unit.declarations.whereType<ClassDeclaration>()) {
      final element = declaration.declaredElement!;
      if (!isAnnotationPresent(element)) continue;
      generationFutures.add(WelltestedSaver().saveForAnnotatedElement(
        element: element,
        annotation: element.metadata.firstWhere(
            (element) => element.element!.displayName == "Welltested"),
        pubspecData: pubspecData,
        dartServerRepository: dartServerRepository,
        classDeclaration: declaration,
        testRunInfo: testRunInfo,
        classFile: file as File,
      ));

      continue;
    }
    await Future.wait(generationFutures);
  } catch (e, stackTrace) {
    wtTelemetry.trackError(
        severity: Severity.error, error: e, stackTrace: stackTrace);
  }
}

runMockitoCommand() async {
  wtLog.startSpinner("Generating mocks");
  await Future.delayed(Duration(seconds: 1));
  final process = await Process.run(
      "dart", ["run", "build_runner", "build", "--delete-conflicting-outputs"],
      runInShell: true);

  if (process.exitCode != 0) {
    wtLog.stopSpinner(
        message: "Failed to generate mocks:",
        severity: MessageSeverity.Warning);
    wtLog.log(process.stderr);
    wtLog.log(
        'Note: You may also directly use the command `dart run build_runner build` to generate the mocks after fixing the above errors.');
  } else {
    wtLog.stopSpinner(
        message: 'Successfully generated mocks.',
        severity: MessageSeverity.Info);
  }
}

Future<Map<String, MethodDeclaration>> generateForElement(
    Element element) async {
  final classVisitor = ClassVisitor();
  final Map<String, MethodDeclaration> methodAstNodes = {};

  element.visitChildren(classVisitor);
  for (final method in classVisitor.methods.entries) {
    final node =
        (await getResolvedAstNodeForElement(method.value)) as MethodDeclaration;
    methodAstNodes[method.key] = node;
  }
  return methodAstNodes;
}

Future<AstNode> getResolvedAstNodeForElement(
  Element element,
) async {
  final resolvedLib =
      await element.session!.getResolvedLibraryByElement(element.library!);
  final astNode = (resolvedLib as ResolvedLibraryResult)
      .getElementDeclaration(element)!
      .node;
  return astNode;
}

==> welltested-2.0.1.tar.gz: lib/cli/operation.dart <==
import 'dart:async';
import 'dart:io';

import 'package:analyzer/dart/analysis/analysis_context_collection.dart';
import 'package:analyzer/dart/ast/ast.dart';
import 'package:analyzer/dart/element/element.dart';
import 'package:analyzer/file_system/physical_file_system.dart';
import 'package:azure_application_insights/azure_application_insights.dart';
import 'package:welltested/generators/saver.dart';
import 'package:welltested/src/extensions/class_element_extension.dart';
import 'package:welltested/src/extensions/file_extensions.dart';
import 'package:welltested/src/models/pubspec_data.dart';
import 'package:welltested/src/models/test_run_info.dart';
import 'package:welltested/src/utilities/cache_manager.dart';
import 'package:welltested/src/utilities/logging.dart';
import 'package:welltested/src/utilities/path_utils.dart';
import 'package:welltested/src/utilities/telemetry_client.dart';
import 'package:welltested/src/utilities/warnings.dart';
import 'package:welltested_annotation/welltested_annotation.dart';

import '../generators/unit_generator.dart';
import '../src/repositories/dart_server_repository.dart';
import '../src/utilities/utilities.dart';
import '../src/visitors/class_visitor.dart';

class CLIOperation {
  static Future<void> saveUnit({required PubspecData pubspecData}) async {
    wtLog.startSpinner("Getting ready...");
    final dartServerRepository = DartServerRepository();
    await dartServerRepository.initialize();

    Iterable<FileSystemEntity> fileList = Directory(PathUtils.libPath)
        .listSync(recursive: true)
        .where((element) => element.path.endsWith(".dart"));

    AnalysisContextCollection collection = AnalysisContextCollection(
      includedPaths: [PathUtils.libPath],
      resourceProvider: PhysicalResourceProvider.INSTANCE,
    );

    List<Future> saveFutures = [];
    final testRunInfo = TestRunInfo.newRun();
    for (var file in fileList) {
      saveFutures.add(
        runSaveUnit(
            file, collection, pubspecData, dartServerRepository, testRunInfo),
      );
    }
    wtLog.startSpinner("Saving Unit Tests");
    await Future.wait(saveFutures);
    wtLog.stopSpinner(
        message:
            '✅ Succeeded. ${testRunInfo.updated} were saved.\nNote: Saving working tests helps improve future test generations. Keep it up!');
    dartServerRepository.dispose();
  }

  static Future<void> generateUnit({
    required PubspecData pubspecData,
    String? filePath,
    String? className,
    String? methodName,
  }) async {
    wtLog.startSpinner("Getting ready...");
    final dartServerRepository = DartServerRepository();
    await dartServerRepository.initialize();

    Iterable<FileSystemEntity> fileList = filePath != null
        ? [File(filePath)]
        : (Directory(PathUtils.libPath).listSync(recursive: true))
            .where((element) => element.path.endsWith(".dart"));

    AnalysisContextCollection collection = AnalysisContextCollection(
      includedPaths: [PathUtils.libPath],
      resourceProvider: PhysicalResourceProvider.INSTANCE,
    );
    List<Future> generationFutures = [];
    final testRunInfo = TestRunInfo.newRun();

    for (var file in fileList) {
      generationFutures.add(
        runUnit(
            file, collection, pubspecData, dartServerRepository, testRunInfo,
            className: className, methodName: methodName),
      );
    }
    wtLog.startSpinner('Generating unit tests');
    await Future.wait(generationFutures);

    wtLog.stopSpinner(
        message:
            '✅ Succeeded. Please find the generated tests in `/test` folder.\nCreated: ${testRunInfo.created}\nUpdated:${testRunInfo.updated}\nFailed:${testRunInfo.failed}');

    dartServerRepository.dispose();
    await runMockitoCommand();
  }
}

Future<void> runUnit(
    FileSystemEntity file,
    AnalysisContextCollection collection,
    PubspecData pubspecData,
    DartServerRepository dartServerRepository,
    TestRunInfo testRunInfo,
    {String? className,
    String? methodName}) async {
  try {
    final resolvedUnit = await file.resolveFile(collection);
    CompilationUnit unit = resolvedUnit.unit;

    final unitGenerator = UnitGenerator(
        dartServerRepository: dartServerRepository,
        cacheManager: CacheManager());
    unitGenerator.init();

    List<Future> generationFutures = [];
    for (final declaration in unit.declarations.whereType<ClassDeclaration>()) {
      final element = declaration.declaredElement!;
      //selection conditions
      if (!isAnnotationPresent(element)) continue;
      if (className != null && declaration.name.toString() != className)
        continue;
      if (element.checkIfAWidgetOrState)
        wtLog.warning(Warnings.annotationOnWidget(element.displayName));

      generationFutures.add(unitGenerator.generateForClass(
        classElement: element,
        annotation: element.metadata.firstWhere((element) =>
            element.element!.displayName ==
            Welltested().runtimeType.toString()),
        pubspecData: pubspecData,
        classAstNode: declaration,
        compilationUnit: unit,
        testRunInfo: testRunInfo,
        classFile: file as File,
        requestedFunction: methodName,
      ));
    }

    for (final declaration
        in unit.declarations.whereType<ExtensionDeclaration>()) {
      final element = declaration.declaredElement!;

      if (!isAnnotationPresent(element)) continue;
      if (className != null && declaration.name.toString() != className)
        continue;
      generationFutures.add(unitGenerator.generateForExtension(
          compilationUnit: unit,
          annotation: element.metadata.firstWhere((element) =>
              element.element!.displayName ==
              Welltested().runtimeType.toString()),
          extAstNode: declaration,
          extElement: element,
          pubspecData: pubspecData,
          extFile: file as File,
          testRunInfo: testRunInfo,
          requestedFunction: methodName));
    }
    for (final declaration
        in unit.declarations.whereType<FunctionDeclaration>()) {
      final element = declaration.declaredElement! as FunctionElement;
      if (!isAnnotationPresent(element)) continue;
      if (methodName != null && declaration.name.toString() != methodName)
        continue;
      generationFutures.add(unitGenerator.generateForGlobalMethod(
        methodElement: element,
        methodAstNode: declaration,
        compilationUnit: unit,
        annotation: element.metadata.firstWhere((element) =>
            element.element!.displayName ==
            Welltested().runtimeType.toString()),
        pubspecData: pubspecData,
        methodFile: file as File,
        testRunInfo: testRunInfo,
      ));
    }

    await Future.wait(generationFutures);
  } catch (e, stackTrace) {
    wtTelemetry.trackError(
        severity: Severity.error, error: e, stackTrace: stackTrace);
    throw (e);
  }
}

Future<void> runSaveUnit(
  FileSystemEntity file,
  AnalysisContextCollection collection,
  PubspecData pubspecData,
  DartServerRepository dartServerRepository,
  TestRunInfo testRunInfo,
) async {
  try {
    final resolvedUnit = await file.resolveFile(collection);
    CompilationUnit unit = resolvedUnit.unit;

    final saver = WelltestedSaver(
        dartServerRepository: dartServerRepository,
        cacheManager: CacheManager());
    saver.init();

    List<Future> saveFutures = [];

    for (final declaration in unit.declarations.whereType<ClassDeclaration>()) {
      final element = declaration.declaredElement!;

      if (!isAnnotationPresent(element)) continue;

      if (element.checkIfAWidgetOrState)
        wtLog.warning(Warnings.annotationOnWidget(element.displayName));

      saveFutures.add(saver.saveForClass(
          classElement: element,
          annotation: element.metadata.firstWhere((element) =>
              element.element!.displayName ==
              Welltested().runtimeType.toString()),
          pubspecData: pubspecData,
          classAstNode: declaration,
          compilationUnit: unit,
          testRunInfo: testRunInfo,
          classFile: file as File));
    }
    for (final declaration
        in unit.declarations.whereType<ExtensionDeclaration>()) {
      final element = declaration.declaredElement!;

      if (!isAnnotationPresent(element)) continue;

      saveFutures.add(saver.saveForExtension(
          compilationUnit: unit,
          annotation: element.metadata.firstWhere((element) =>
              element.element!.displayName ==
              Welltested().runtimeType.toString()),
          extAstNode: declaration,
          extElement: element,
          pubspecData: pubspecData,
          extFile: file as File,
          testRunInfo: testRunInfo));
    }
    for (final declaration
        in unit.declarations.whereType<FunctionDeclaration>()) {
      final element = declaration.declaredElement! as FunctionElement;
      if (!isAnnotationPresent(element)) continue;

      saveFutures.add(saver.saveForGlobalMethod(
        methodElement: element,
        methodAstNode: declaration,
        compilationUnit: unit,
        annotation: element.metadata.firstWhere((element) =>
            element.element!.displayName ==
            Welltested().runtimeType.toString()),
        pubspecData: pubspecData,
        methodFile: file as File,
        testRunInfo: testRunInfo,
      ));
    }
    await Future.wait(saveFutures);
  } catch (e, stackTrace) {
    wtTelemetry.trackError(
        severity: Severity.error, error: e, stackTrace: stackTrace);
    throw (e);
  }
}

runMockitoCommand() async {
  wtLog.startSpinner("Generating mocks");
  await Future.delayed(Duration(seconds: 1));
  final process = await Process.run(
      "dart", ["run", "build_runner", "build", "--delete-conflicting-outputs"],
      runInShell: true);

  if (process.exitCode != 0) {
    wtLog.stopSpinner(
        message: "Failed to generate mocks:",
        severity: MessageSeverity.Warning);
    wtLog.log(process.stderr);
    wtLog.log(
        'Note: You may also directly use the command `dart run build_runner build` to generate the mocks after fixing the above errors.');
  } else {
    wtLog.stopSpinner(
        message: 'Successfully generated mocks.',
        severity: MessageSeverity.Info);
  }
}

Future<Map<String, MethodDeclaration>> generateForElement(
    Element element) async {
  final classVisitor = ClassVisitor();
  final Map<String, MethodDeclaration> methodAstNodes = {};

  element.visitChildren(classVisitor);
  for (final method in classVisitor.methods.entries) {
    final node =
        (await getResolvedAstNodeForElement(method.value)) as MethodDeclaration;
    methodAstNodes[method.key] = node;
  }
  return methodAstNodes;
}
